/* 
 * Copyright (C) 2013-2014 Pier Luigi Fiorini
 * 
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License as published by
 * the Free Software Foundation, either version 2.1 of the License, or
 * (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Lesser General Public License for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

#ifndef PLASMA_SHELL_SERVER_PROTOCOL_H
#define PLASMA_SHELL_SERVER_PROTOCOL_H

#ifdef  __cplusplus
extern "C" {
#endif

#include <stdint.h>
#include <stddef.h>
#include "wayland-server.h"

struct wl_client;
struct wl_resource;

struct org_kde_plasma_shell;
struct org_kde_plasma_surface;
struct wl_output;
struct wl_surface;

extern const struct wl_interface org_kde_plasma_shell_interface;
extern const struct wl_interface org_kde_plasma_surface_interface;

/**
 * org_kde_plasma_shell - create shell windows and helpers
 * @get_surface: create a shell surface from a surface
 * @desktop_ready: desktop is ready to be shown
 * @present: present all surfaces specified
 *
 * This interface is used by KF5 powered Wayland shells to communicate
 * with the compositor and can only be bound one time.
 */
struct org_kde_plasma_shell_interface {
	/**
	 * get_surface - create a shell surface from a surface
	 * @id: (none)
	 * @surface: (none)
	 *
	 * Create a shell surface for an existing surface.
	 *
	 * Only one shell surface can be associated with a given surface.
	 */
	void (*get_surface)(struct wl_client *client,
			    struct wl_resource *resource,
			    uint32_t id,
			    struct wl_resource *surface);
	/**
	 * desktop_ready - desktop is ready to be shown
	 *
	 * Tell the compositor, that enough desktop elements have been
	 * drawn to make the desktop look ready for use.
	 *
	 * During start-up, the compositor can wait for this request with a
	 * black screen and fade the desktop in when this request is
	 * called.
	 *
	 * If the shell surfaces take a long time to initialize, we avoid
	 * showing temporary garbage.
	 *
	 * See also the splash role in org_kde_plasma_surface.set_role.
	 */
	void (*desktop_ready)(struct wl_client *client,
			      struct wl_resource *resource);
	/**
	 * present - present all surfaces specified
	 * @surfaces: list of surfaces to present
	 *
	 * Tell the compositor to present all the specified surfaces.
	 */
	void (*present)(struct wl_client *client,
			struct wl_resource *resource,
			struct wl_array *surfaces);
};

#define ORG_KDE_PLASMA_SHELL_LOADED	0

#define ORG_KDE_PLASMA_SHELL_LOADED_SINCE_VERSION	1

static inline void
org_kde_plasma_shell_send_loaded(struct wl_resource *resource_)
{
	wl_resource_post_event(resource_, ORG_KDE_PLASMA_SHELL_LOADED);
}

#ifndef ORG_KDE_PLASMA_SURFACE_ROLE_ENUM
#define ORG_KDE_PLASMA_SURFACE_ROLE_ENUM
enum org_kde_plasma_surface_role {
	ORG_KDE_PLASMA_SURFACE_ROLE_NONE = 0,
	ORG_KDE_PLASMA_SURFACE_ROLE_SPLASH = 1,
	ORG_KDE_PLASMA_SURFACE_ROLE_DESKTOP = 2,
	ORG_KDE_PLASMA_SURFACE_ROLE_DASHBOARD = 3,
	ORG_KDE_PLASMA_SURFACE_ROLE_PANEL = 4,
	ORG_KDE_PLASMA_SURFACE_ROLE_OVERLAY = 5,
	ORG_KDE_PLASMA_SURFACE_ROLE_NOTIFICATION = 6,
	ORG_KDE_PLASMA_SURFACE_ROLE_LOCK = 7,
};
#endif /* ORG_KDE_PLASMA_SURFACE_ROLE_ENUM */

#ifndef ORG_KDE_PLASMA_SURFACE_FLAG_ENUM
#define ORG_KDE_PLASMA_SURFACE_FLAG_ENUM
/**
 * org_kde_plasma_surface_flag - flags on the surface
 * @ORG_KDE_PLASMA_SURFACE_FLAG_PANEL_ALWAYS_VISIBLE: normal panel
 *	visibility
 * @ORG_KDE_PLASMA_SURFACE_FLAG_PANEL_AUTO_HIDE: hide automatically
 * @ORG_KDE_PLASMA_SURFACE_FLAG_PANEL_WINDOWS_CAN_COVER: windows can
 *	cover
 * @ORG_KDE_PLASMA_SURFACE_FLAG_PANEL_WINDOWS_GO_BELOW: windows go below
 *
 * Various flags that modify the behavior of the surface. Flags may
 * depend on the surface role.
 *
 * Ranges are:
 *
 * 0x0000 - 0x007F: panel visibility
 */
enum org_kde_plasma_surface_flag {
	ORG_KDE_PLASMA_SURFACE_FLAG_PANEL_ALWAYS_VISIBLE = 1,
	ORG_KDE_PLASMA_SURFACE_FLAG_PANEL_AUTO_HIDE = 2,
	ORG_KDE_PLASMA_SURFACE_FLAG_PANEL_WINDOWS_CAN_COVER = 3,
	ORG_KDE_PLASMA_SURFACE_FLAG_PANEL_WINDOWS_GO_BELOW = 4,
};
#endif /* ORG_KDE_PLASMA_SURFACE_FLAG_ENUM */

#ifndef ORG_KDE_PLASMA_SURFACE_SCREEN_EDGE_ENUM
#define ORG_KDE_PLASMA_SURFACE_SCREEN_EDGE_ENUM
enum org_kde_plasma_surface_screen_edge {
	ORG_KDE_PLASMA_SURFACE_SCREEN_EDGE_NONE = 0,
	ORG_KDE_PLASMA_SURFACE_SCREEN_EDGE_LEFT = 1,
	ORG_KDE_PLASMA_SURFACE_SCREEN_EDGE_TOP = 2,
	ORG_KDE_PLASMA_SURFACE_SCREEN_EDGE_RIGHT = 3,
	ORG_KDE_PLASMA_SURFACE_SCREEN_EDGE_BOTTOM = 4,
};
#endif /* ORG_KDE_PLASMA_SURFACE_SCREEN_EDGE_ENUM */

/**
 * org_kde_plasma_surface - metadata interface
 * @destroy: remove org_kde_plasma_surface interface
 * @set_output: assign an output to this shell surface
 * @set_position: change the shell surface position
 * @set_role: assign a role to this surface
 * @set_flags: set or unset flags specific to the role
 * @set_screen_edge: set screen edge activation
 *
 * An interface that may be implemented by a wl_surface, for
 * implementations that provide the shell user interface.
 *
 * It provides requests to set surface roles, assign an output or set the
 * position in output coordinates.
 *
 * On the server side the object is automatically destroyed when the
 * related wl_surface is destroyed. On client side,
 * org_kde_plasma_surface.destroy() must be called before destroying the
 * wl_surface object.
 */
struct org_kde_plasma_surface_interface {
	/**
	 * destroy - remove org_kde_plasma_surface interface
	 *
	 * The org_kde_plasma_surface interface is removed from the
	 * wl_surface object that was turned into a shell surface with the
	 * org_kde_plasma_shell.get_surface request. The shell surface role
	 * is lost and wl_surface is unmapped.
	 */
	void (*destroy)(struct wl_client *client,
			struct wl_resource *resource);
	/**
	 * set_output - assign an output to this shell surface
	 * @output: (none)
	 *
	 * Assign an output to this shell surface. The compositor will
	 * use this information to set the position when
	 * org_kde_plasma_surface.set_position request is called.
	 */
	void (*set_output)(struct wl_client *client,
			   struct wl_resource *resource,
			   struct wl_resource *output);
	/**
	 * set_position - change the shell surface position
	 * @x: x coordinate in global space
	 * @y: y coordinate in global space
	 *
	 * Move the surface to new coordinates.
	 *
	 * Coordinates are global, for example 50,50 for a 1920,0+1920x1080
	 * output is 1970,50 in global coordinates space.
	 *
	 * Use org_kde_plasma_surface.set_output to assign an output to
	 * this surface.
	 */
	void (*set_position)(struct wl_client *client,
			     struct wl_resource *resource,
			     int32_t x,
			     int32_t y);
	/**
	 * set_role - assign a role to this surface
	 * @role: (none)
	 *
	 * Assign a role to a shell surface.
	 *
	 * The compositor handles surfaces depending on their role. See the
	 * explaination below.
	 *
	 * This request fails if the surface already has a role, this means
	 * the surface role may be assigned only once.
	 *
	 * == Surfaces with splash role ==
	 *
	 * Splash surfaces are placed above every other surface during the
	 * shell startup phase.
	 *
	 * The surfaces are placed according to the output coordinates. No
	 * size is imposed to those surfaces, the shell has to resize them
	 * according to output size.
	 *
	 * These surfaces are meant to hide the desktop during the startup
	 * phase so that the user will always see a ready to work desktop.
	 *
	 * A shell might not create splash surfaces if the compositor
	 * reveals the desktop in an alternative fashion, for example with
	 * a fade in effect.
	 *
	 * That depends on how much time the desktop usually need to
	 * prepare the workspace or specific design decisions. This
	 * specification doesn't impose any particular design.
	 *
	 * When the startup phase is finished, the shell will send the
	 * org_kde_plasma.desktop_ready request to the compositor.
	 *
	 * == Surfaces with desktop role ==
	 *
	 * Desktop surfaces are placed below all other surfaces and are
	 * used to show the actual desktop view with icons, search results
	 * or controls the user will interact with. What to show depends on
	 * the shell implementation.
	 *
	 * The surfaces are placed according to the output coordinates. No
	 * size is imposed to those surfaces, the shell has to resize them
	 * according to output size.
	 *
	 * Only one surface per output can have the desktop role.
	 *
	 * == Surfaces with dashboard role ==
	 *
	 * Dashboard surfaces are placed above desktop surfaces and are
	 * used to show additional widgets and controls.
	 *
	 * The surfaces are placed according to the output coordinates. No
	 * size is imposed to those surfaces, the shell has to resize them
	 * according to output size.
	 *
	 * Only one surface per output can have the dashboard role.
	 *
	 * == Surfaces with config role ==
	 *
	 * A configuration surface is shown when the user wants to
	 * configure panel or desktop views.
	 *
	 * Only one surface per output can have the config role.
	 *
	 * TODO: This should grab the input like popup menus, right?
	 *
	 * == Surfaces with overlay role ==
	 *
	 * Overlays are special surfaces that shows for a limited amount of
	 * time. Such surfaces are useful to display things like volume,
	 * brightness and status changes.
	 *
	 * Compositors may decide to show those surfaces in a layer above
	 * all surfaces, even full screen ones if so is desired.
	 *
	 * == Surfaces with notification role ==
	 *
	 * Notification surfaces display informative content for a limited
	 * amount of time. The compositor may decide to show them in a
	 * corner depending on the configuration.
	 *
	 * These surfaces are shown in a layer above all other surfaces
	 * except for full screen ones.
	 *
	 * == Surfaces with lock role ==
	 *
	 * The lock surface is shown by the compositor when the session is
	 * locked, users interact with it to unlock the session.
	 *
	 * Compositors should move lock surfaces to 0,0 in output
	 * coordinates space and hide all other surfaces for security sake.
	 * For the same reason it is recommended that clients make the lock
	 * surface as big as the screen.
	 *
	 * Only one surface per output can have the lock role.
	 */
	void (*set_role)(struct wl_client *client,
			 struct wl_resource *resource,
			 uint32_t role);
	/**
	 * set_flags - set or unset flags specific to the role
	 * @flags: bitmask of flag values
	 *
	 * Set flags bitmask as described by the flag enum. Pass 0 to
	 * unset any flag, the surface will adjust its behavior to the
	 * default.
	 */
	void (*set_flags)(struct wl_client *client,
			  struct wl_resource *resource,
			  uint32_t flags);
	/**
	 * set_screen_edge - set screen edge activation
	 * @screen_edge: (none)
	 *
	 * 
	 */
	void (*set_screen_edge)(struct wl_client *client,
				struct wl_resource *resource,
				uint32_t screen_edge);
};


#ifdef  __cplusplus
}
#endif

#endif
