/* 
 * Copyright (C) 2014 Jolla Ltd.
 * 
 * Permission to use, copy, modify, distribute, and sell this
 * software and its documentation for any purpose is hereby granted
 * without fee, provided that the above copyright notice appear in
 * all copies and that both that copyright notice and this permission
 * notice appear in supporting documentation, and that the name of
 * the copyright holders not be used in advertising or publicity
 * pertaining to distribution of the software without specific,
 * written prior permission.  The copyright holders make no
 * representations about the suitability of this software for any
 * purpose.  It is provided "as is" without express or implied
 * warranty.
 * 
 * THE COPYRIGHT HOLDERS DISCLAIM ALL WARRANTIES WITH REGARD TO THIS
 * SOFTWARE, INCLUDING ALL IMPLIED WARRANTIES OF MERCHANTABILITY AND
 * FITNESS, IN NO EVENT SHALL THE COPYRIGHT HOLDERS BE LIABLE FOR ANY
 * SPECIAL, INDIRECT OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
 * WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN
 * AN ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION,
 * ARISING OUT OF OR IN CONNECTION WITH THE USE OR PERFORMANCE OF
 * THIS SOFTWARE.
 */

#ifndef GREENISLAND_RECORDER_SERVER_PROTOCOL_H
#define GREENISLAND_RECORDER_SERVER_PROTOCOL_H

#ifdef  __cplusplus
extern "C" {
#endif

#include <stdint.h>
#include <stddef.h>
#include "wayland-server.h"

struct wl_client;
struct wl_resource;

struct greenisland_recorder;
struct greenisland_recorder_manager;
struct wl_buffer;
struct wl_output;

extern const struct wl_interface greenisland_recorder_manager_interface;
extern const struct wl_interface greenisland_recorder_interface;

struct greenisland_recorder_manager_interface {
	/**
	 * create_recorder - create a recorder object
	 * @recorder: (none)
	 * @output: (none)
	 *
	 * Create a recorder object for the specified output.
	 */
	void (*create_recorder)(struct wl_client *client,
				struct wl_resource *resource,
				uint32_t recorder,
				struct wl_resource *output);
};


#ifndef GREENISLAND_RECORDER_RESULT_ENUM
#define GREENISLAND_RECORDER_RESULT_ENUM
enum greenisland_recorder_result {
	GREENISLAND_RECORDER_RESULT_BAD_BUFFER = 2,
};
#endif /* GREENISLAND_RECORDER_RESULT_ENUM */

#ifndef GREENISLAND_RECORDER_TRANSFORM_ENUM
#define GREENISLAND_RECORDER_TRANSFORM_ENUM
enum greenisland_recorder_transform {
	GREENISLAND_RECORDER_TRANSFORM_NORMAL = 1,
	GREENISLAND_RECORDER_TRANSFORM_Y_INVERTED = 2,
};
#endif /* GREENISLAND_RECORDER_TRANSFORM_ENUM */

struct greenisland_recorder_interface {
	/**
	 * destroy - destroy the recorder object
	 *
	 * Destroy the recorder object, discarding any frame request that
	 * may be pending.
	 */
	void (*destroy)(struct wl_client *client,
			struct wl_resource *resource);
	/**
	 * record_frame - request a frame to be recorded
	 * @buffer: (none)
	 *
	 * Ask the compositor to record its next frame, putting the
	 * content into the specified buffer data. The frame event will be
	 * sent when the frame is recorded. Only one frame will be
	 * recorded, the client will have to call this again after the
	 * frame event if it wants to record more frames.
	 *
	 * The buffer must be a shm buffer, trying to use another type of
	 * buffer will result in failure to capture the frame and the
	 * failed event will be sent.
	 */
	void (*record_frame)(struct wl_client *client,
			     struct wl_resource *resource,
			     struct wl_resource *buffer);
	/**
	 * repaint - request the compositor to repaint asap
	 *
	 * Calling record_frame will not cause the compositor to repaint,
	 * but it will wait instead for the first frame the compositor
	 * draws due to some other external event or internal change.
	 * Calling this request after calling record_frame will ask the
	 * compositor to redraw as soon at possible even if it wouldn't
	 * otherwise. If no frame was requested this request has no effect.
	 */
	void (*repaint)(struct wl_client *client,
			struct wl_resource *resource);
};

#define GREENISLAND_RECORDER_SETUP	0
#define GREENISLAND_RECORDER_FRAME	1
#define GREENISLAND_RECORDER_FAILED	2
#define GREENISLAND_RECORDER_CANCELLED	3

#define GREENISLAND_RECORDER_SETUP_SINCE_VERSION	1
#define GREENISLAND_RECORDER_FRAME_SINCE_VERSION	1
#define GREENISLAND_RECORDER_FAILED_SINCE_VERSION	1
#define GREENISLAND_RECORDER_CANCELLED_SINCE_VERSION	1

static inline void
greenisland_recorder_send_setup(struct wl_resource *resource_, int32_t width, int32_t height, int32_t stride, int32_t format)
{
	wl_resource_post_event(resource_, GREENISLAND_RECORDER_SETUP, width, height, stride, format);
}

static inline void
greenisland_recorder_send_frame(struct wl_resource *resource_, struct wl_resource *buffer, uint32_t time, int32_t transform)
{
	wl_resource_post_event(resource_, GREENISLAND_RECORDER_FRAME, buffer, time, transform);
}

static inline void
greenisland_recorder_send_failed(struct wl_resource *resource_, int32_t result, struct wl_resource *buffer)
{
	wl_resource_post_event(resource_, GREENISLAND_RECORDER_FAILED, result, buffer);
}

static inline void
greenisland_recorder_send_cancelled(struct wl_resource *resource_, struct wl_resource *buffer)
{
	wl_resource_post_event(resource_, GREENISLAND_RECORDER_CANCELLED, buffer);
}

#ifdef  __cplusplus
}
#endif

#endif
