/* 
 * Copyright (C) 2014 Jolla Ltd.
 * 
 * Permission to use, copy, modify, distribute, and sell this
 * software and its documentation for any purpose is hereby granted
 * without fee, provided that the above copyright notice appear in
 * all copies and that both that copyright notice and this permission
 * notice appear in supporting documentation, and that the name of
 * the copyright holders not be used in advertising or publicity
 * pertaining to distribution of the software without specific,
 * written prior permission.  The copyright holders make no
 * representations about the suitability of this software for any
 * purpose.  It is provided "as is" without express or implied
 * warranty.
 * 
 * THE COPYRIGHT HOLDERS DISCLAIM ALL WARRANTIES WITH REGARD TO THIS
 * SOFTWARE, INCLUDING ALL IMPLIED WARRANTIES OF MERCHANTABILITY AND
 * FITNESS, IN NO EVENT SHALL THE COPYRIGHT HOLDERS BE LIABLE FOR ANY
 * SPECIAL, INDIRECT OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
 * WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN
 * AN ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION,
 * ARISING OUT OF OR IN CONNECTION WITH THE USE OR PERFORMANCE OF
 * THIS SOFTWARE.
 */

#ifndef GREENISLAND_RECORDER_CLIENT_PROTOCOL_H
#define GREENISLAND_RECORDER_CLIENT_PROTOCOL_H

#ifdef  __cplusplus
extern "C" {
#endif

#include <stdint.h>
#include <stddef.h>
#include "wayland-client.h"

struct wl_client;
struct wl_resource;

struct greenisland_recorder;
struct greenisland_recorder_manager;
struct wl_buffer;
struct wl_output;

extern const struct wl_interface greenisland_recorder_manager_interface;
extern const struct wl_interface greenisland_recorder_interface;

#define GREENISLAND_RECORDER_MANAGER_CREATE_RECORDER	0

static inline void
greenisland_recorder_manager_set_user_data(struct greenisland_recorder_manager *greenisland_recorder_manager, void *user_data)
{
	wl_proxy_set_user_data((struct wl_proxy *) greenisland_recorder_manager, user_data);
}

static inline void *
greenisland_recorder_manager_get_user_data(struct greenisland_recorder_manager *greenisland_recorder_manager)
{
	return wl_proxy_get_user_data((struct wl_proxy *) greenisland_recorder_manager);
}

static inline void
greenisland_recorder_manager_destroy(struct greenisland_recorder_manager *greenisland_recorder_manager)
{
	wl_proxy_destroy((struct wl_proxy *) greenisland_recorder_manager);
}

static inline struct greenisland_recorder *
greenisland_recorder_manager_create_recorder(struct greenisland_recorder_manager *greenisland_recorder_manager, struct wl_output *output)
{
	struct wl_proxy *recorder;

	recorder = wl_proxy_marshal_constructor((struct wl_proxy *) greenisland_recorder_manager,
			 GREENISLAND_RECORDER_MANAGER_CREATE_RECORDER, &greenisland_recorder_interface, NULL, output);

	return (struct greenisland_recorder *) recorder;
}

#ifndef GREENISLAND_RECORDER_RESULT_ENUM
#define GREENISLAND_RECORDER_RESULT_ENUM
enum greenisland_recorder_result {
	GREENISLAND_RECORDER_RESULT_BAD_BUFFER = 2,
};
#endif /* GREENISLAND_RECORDER_RESULT_ENUM */

#ifndef GREENISLAND_RECORDER_TRANSFORM_ENUM
#define GREENISLAND_RECORDER_TRANSFORM_ENUM
enum greenisland_recorder_transform {
	GREENISLAND_RECORDER_TRANSFORM_NORMAL = 1,
	GREENISLAND_RECORDER_TRANSFORM_Y_INVERTED = 2,
};
#endif /* GREENISLAND_RECORDER_TRANSFORM_ENUM */

struct greenisland_recorder_listener {
	/**
	 * setup - notify the requirements for the frame buffers
	 * @width: (none)
	 * @height: (none)
	 * @stride: (none)
	 * @format: (none)
	 *
	 * This event will be sent immediately after creation of the
	 * greenisland_recorder object. The wl_buffers the client passes to
	 * the frame request must be big enough to store an image with the
	 * given width, height and format. If they are not the compositor
	 * will send the failed event. If this event is sent again later in
	 * the lifetime of the object the pending frames will be cancelled.
	 *
	 * The format will be one of the values as defined in the
	 * wl_shm::format enum.
	 */
	void (*setup)(void *data,
		      struct greenisland_recorder *greenisland_recorder,
		      int32_t width,
		      int32_t height,
		      int32_t stride,
		      int32_t format);
	/**
	 * frame - notify a frame was recorded, or an error
	 * @buffer: (none)
	 * @time: (none)
	 * @transform: (none)
	 *
	 * The compositor will send this event after a frame was
	 * recorded, or in case an error happened. The client can call
	 * record_frame again to record the next frame.
	 *
	 * 'time' is the time the compositor recorded that frame, in
	 * milliseconds, with an unspecified base.
	 */
	void (*frame)(void *data,
		      struct greenisland_recorder *greenisland_recorder,
		      struct wl_buffer *buffer,
		      uint32_t time,
		      int32_t transform);
	/**
	 * failed - the frame capture failed
	 * @result: (none)
	 * @buffer: (none)
	 *
	 * The value of the 'result' argument will be one of the values
	 * of the 'result' enum.
	 */
	void (*failed)(void *data,
		       struct greenisland_recorder *greenisland_recorder,
		       int32_t result,
		       struct wl_buffer *buffer);
	/**
	 * cancelled - notify a request was cancelled
	 * @buffer: (none)
	 *
	 * The compositor will send this event if the client calls
	 * request_frame more than one time for the same compositor frame.
	 * The cancel event will be sent carrying the old buffer, and the
	 * frame will be recorded using the newest buffer.
	 */
	void (*cancelled)(void *data,
			  struct greenisland_recorder *greenisland_recorder,
			  struct wl_buffer *buffer);
};

static inline int
greenisland_recorder_add_listener(struct greenisland_recorder *greenisland_recorder,
				  const struct greenisland_recorder_listener *listener, void *data)
{
	return wl_proxy_add_listener((struct wl_proxy *) greenisland_recorder,
				     (void (**)(void)) listener, data);
}

#define GREENISLAND_RECORDER_DESTROY	0
#define GREENISLAND_RECORDER_RECORD_FRAME	1
#define GREENISLAND_RECORDER_REPAINT	2

static inline void
greenisland_recorder_set_user_data(struct greenisland_recorder *greenisland_recorder, void *user_data)
{
	wl_proxy_set_user_data((struct wl_proxy *) greenisland_recorder, user_data);
}

static inline void *
greenisland_recorder_get_user_data(struct greenisland_recorder *greenisland_recorder)
{
	return wl_proxy_get_user_data((struct wl_proxy *) greenisland_recorder);
}

static inline void
greenisland_recorder_destroy(struct greenisland_recorder *greenisland_recorder)
{
	wl_proxy_marshal((struct wl_proxy *) greenisland_recorder,
			 GREENISLAND_RECORDER_DESTROY);

	wl_proxy_destroy((struct wl_proxy *) greenisland_recorder);
}

static inline void
greenisland_recorder_record_frame(struct greenisland_recorder *greenisland_recorder, struct wl_buffer *buffer)
{
	wl_proxy_marshal((struct wl_proxy *) greenisland_recorder,
			 GREENISLAND_RECORDER_RECORD_FRAME, buffer);
}

static inline void
greenisland_recorder_repaint(struct greenisland_recorder *greenisland_recorder)
{
	wl_proxy_marshal((struct wl_proxy *) greenisland_recorder,
			 GREENISLAND_RECORDER_REPAINT);
}

#ifdef  __cplusplus
}
#endif

#endif
